// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:bloc/bloc.dart';
import 'package:equatable/equatable.dart';
import 'package:wyatt_crud_bloc/src/models/model.dart';
import 'package:wyatt_crud_bloc/src/models/queries/queries_interface.dart';
import 'package:wyatt_crud_bloc/src/repositories/crud_repository_interface.dart';

part 'crud_state.dart';

class CrudCubit<T extends Model<Object?, T>> extends Cubit<CrudState<T>> {
  final CrudRepositoryInterface<T> _crudRepository;

  // ignore: prefer_const_constructors
  CrudCubit(this._crudRepository) : super(CrudState());
  // Here we can't use `const` because we need the generic type T

  void reset() {
    // ignore: prefer_const_constructors
    emit(CrudState());
    // Same here, because of `const` we can't use T generic type
  }

  Future<void> create(Model object, {String? id}) async {
    emit(state.copyWith(status: CrudStatus.loading));
    try {
      await _crudRepository.create(object, id: id);
      final data = state.data..addAll([object].cast<T>());
      emit(
        state.copyWith(
          data: data,
          status: CrudStatus.success,
        ),
      );
    } catch (e) {
      // TODO(hpcl): implement Exception
      emit(
        state.copyWith(
          status: CrudStatus.failure,
          errorMessage: e.toString(),
        ),
      );
    }
  }

  Future<void> delete(String id) async {
    emit(state.copyWith(status: CrudStatus.loading));
    try {
      await _crudRepository.delete(id);
      final data = state.data
        ..removeWhere((element) => element!.id != null && element.id == id);
      emit(
        state.copyWith(
          data: data,
          status: CrudStatus.success,
        ),
      );
    } catch (e) {
      emit(
        state.copyWith(
          status: CrudStatus.failure,
          errorMessage: e.toString(),
        ),
      );
    }
  }

  Future<void> deleteAll() async {
    emit(state.copyWith(status: CrudStatus.loading));
    try {
      await _crudRepository.deleteAll();
      emit(
        state.copyWith(
          data: [],
          status: CrudStatus.success,
        ),
      );
    } catch (e) {
      emit(
        state.copyWith(
          status: CrudStatus.failure,
          errorMessage: e.toString(),
        ),
      );
    }
  }

  Future<void> get(String id) async {
    emit(state.copyWith(status: CrudStatus.loading));
    try {
      final data = await _crudRepository.get(id);
      emit(
        state.copyWith(
          data: [data].cast<T>(),
          status: CrudStatus.success,
        ),
      );
    } catch (e) {
      emit(
        state.copyWith(
          status: CrudStatus.failure,
          errorMessage: e.toString(),
        ),
      );
    }
  }

  Future<void> getAll() async {
    emit(state.copyWith(status: CrudStatus.loading));
    try {
      final data = await _crudRepository.getAll();
      emit(
        state.copyWith(
          data: data.cast<T>(),
          status: CrudStatus.success,
        ),
      );
    } catch (e) {
      emit(
        state.copyWith(
          status: CrudStatus.failure,
          errorMessage: e.toString(),
        ),
      );
    }
  }

  Future<void> query(List<QueryInterface> conditions) async {
    emit(state.copyWith(status: CrudStatus.loading));
    try {
      final data = await _crudRepository.query(conditions);
      emit(
        state.copyWith(
          data: data.cast<T>(),
          status: CrudStatus.success,
        ),
      );
    } catch (e) {
      emit(
        state.copyWith(
          status: CrudStatus.failure,
          errorMessage: e.toString(),
        ),
      );
    }
  }

  Future<void> streamOf({String? id}) async {
    emit(state.copyWith(status: CrudStatus.loading));
    try {
      final Stream<List<T?>> data = _crudRepository.stream(id: id);
      emit(
        state.copyWith(
          stream: data,
          status: CrudStatus.success,
        ),
      );
    } catch (e) {
      emit(
        state.copyWith(
          status: CrudStatus.failure,
          errorMessage: e.toString(),
        ),
      );
    }
  }

  Future<void> update(
    String id, {
    Model? object,
    Map<String, dynamic>? raw,
  }) async {
    emit(state.copyWith(status: CrudStatus.loading));
    try {
      await _crudRepository.update(
        id,
        object: object,
        raw: raw,
      );
      emit(state.copyWith(status: CrudStatus.success));
    } catch (e) {
      emit(
        state.copyWith(
          status: CrudStatus.failure,
          errorMessage: e.toString(),
        ),
      );
    }
  }

  Future<void> updateAll(Map<String, Object?> raw) async {
    emit(state.copyWith(status: CrudStatus.loading));
    try {
      await _crudRepository.updateAll(raw);
      emit(state.copyWith(status: CrudStatus.success));
    } catch (e) {
      emit(
        state.copyWith(
          status: CrudStatus.failure,
          errorMessage: e.toString(),
        ),
      );
    }
  }
}
